@@ -1,9 +1,11 @@
 /* utility functions for `patch' */
 
-/* $Id: util.c,v 1.27 1999/08/30 06:20:08 eggert Exp $ */
+/* $Id: util.c,v 1.36 2003/05/20 14:04:53 eggert Exp $ */
 
-/* Copyright 1986 Larry Wall
-   Copyright 1992, 1993, 1997-1998, 1999 Free Software Foundation, Inc.
+/* Copyright (C) 1986 Larry Wall
+
+   Copyright (C) 1992, 1993, 1997, 1998, 1999, 2001, 2002, 2003 Free
+   Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -23,7 +25,7 @@
 #define XTERN extern
 #include <common.h>
 #include <backupfile.h>
-#include <basename.h>
+#include <dirname.h>
 #include <quotearg.h>
 #include <quotesys.h>
 #include <version.h>
@@ -43,23 +45,17 @@
 # define raise(sig) kill (getpid (), sig)
 #endif
 
-#ifdef __STDC__
-# include <stdarg.h>
-# define vararg_start va_start
-#else
-# define vararg_start(ap,p) va_start (ap)
-# if HAVE_VARARGS_H
-#  include <varargs.h>
-# else
-   typedef char *va_list;
-#  define va_dcl int va_alist;
-#  define va_start(ap) ((ap) = (va_list) &va_alist)
-#  define va_arg(ap, t) (((t *) ((ap) += sizeof (t)))  [-1])
-#  define va_end(ap)
-# endif
+#if defined(HAVE_STAT_TIMEVAL)
+#include <time.h>
 #endif
 
-static void makedirs PARAMS ((char *));
+#include <stdarg.h>
+#include <hash.h>
+
+static void makedirs (char *);
+static bool fid_search (const char *, const struct stat *, bool);
+# define fid_exists(name, pst) fid_search (name, pst, false)
+# define insert_fid(name) fid_search (name, NULL, true)
 
 /* Move a file FROM (where *FROM_NEEDS_REMOVAL is nonzero if FROM
    needs removal when cleaning up at the end of execution)
@@ -67,16 +63,16 @@ static void makedirs PARAMS ((char *));
    If we must create TO, use MODE to create it.
    If FROM is null, remove TO (ignoring FROMSTAT).
    FROM_NEEDS_REMOVAL must be nonnull if FROM is nonnull.
-   Back up TO if BACKUP is nonzero.  */
+   Back up TO if BACKUP is true.  */
 
 void
 move_file (char const *from, int volatile *from_needs_removal,
-	   char *to, mode_t mode, int backup)
+	   char *to, mode_t mode, bool backup)
 {
   struct stat to_st;
   int to_errno = ! backup ? -1 : stat (to, &to_st) == 0 ? 0 : errno;
 
-  if (backup)
+  if (backup && (to_errno || ! fid_exists (to, &to_st)))
     {
       int try_makedirs_errno = 0;
       char *bakname;
@@ -152,11 +148,9 @@ move_file (char const *from, int volatile *from_needs_removal,
 	say ("Renaming file %s to %s\n",
 	     quotearg_n (0, from), quotearg_n (1, to));
 
-      if (rename (from, to) == 0)
-	*from_needs_removal = 0;
-      else
+      if (rename (from, to) != 0)
 	{
-	  int to_dir_known_to_exist = 0;
+	  bool to_dir_known_to_exist = false;
 
 	  if (errno == ENOENT
 	      && (to_errno == -1 || to_errno == ENOENT))
@@ -164,10 +158,7 @@ move_file (char const *from, int volatile *from_needs_removal,
 	      makedirs (to);
 	      to_dir_known_to_exist = 1;
 	      if (rename (from, to) == 0)
-		{
-		  *from_needs_removal = 0;
-		  return;
-		}
+		goto rename_succeeded;
 	    }
 
 	  if (errno == EXDEV)
@@ -175,19 +166,29 @@ move_file (char const *from, int volatile *from_needs_removal,
 	      if (! backup)
 		{
 		  if (unlink (to) == 0)
-		    to_dir_known_to_exist = 1;
+		    to_dir_known_to_exist = true;
 		  else if (errno != ENOENT)
 		    pfatal ("Can't remove file %s", quotearg (to));
 		}
 	      if (! to_dir_known_to_exist)
 		makedirs (to);
 	      copy_file (from, to, 0, mode);
+	      insert_fid (to);
 	      return;
 	    }
 
 	  pfatal ("Can't rename file %s to %s",
 		  quotearg_n (0, from), quotearg_n (1, to));
 	}
+
+    rename_succeeded:
+      insert_fid (to);
+      /* Do not clear *FROM_NEEDS_REMOVAL if it's possible that the
+	 rename returned zero because FROM and TO are hard links to
+	 the same file.  */
+      if (0 < to_errno
+	  || (to_errno == 0 && to_st.st_nlink <= 1))
+	*from_needs_removal = 0;
     }
   else if (! backup)
     {
@@ -255,23 +256,27 @@ static char const SCCSDIFF2[] = "|diff - %s";
 
 static char const CLEARTOOL_CO[] = "cleartool co -unr -nc ";
 
+static char const PERFORCE_CO[] = "p4 edit ";
+
 /* Return "RCS" if FILENAME is controlled by RCS,
    "SCCS" if it is controlled by SCCS,
-   "ClearCase" if it is controlled by Clearcase, and 0 otherwise.
-   READONLY is nonzero if we desire only readonly access to FILENAME.
+   "ClearCase" if it is controlled by Clearcase,
+   "Perforce" if it is controlled by Perforce,
+   and 0 otherwise.
+   READONLY is true if we desire only readonly access to FILENAME.
    FILESTAT describes FILENAME's status or is 0 if FILENAME does not exist.
    If successful and if GETBUF is nonzero, set *GETBUF to a command
    that gets the file; similarly for DIFFBUF and a command to diff the file
    (but set *DIFFBUF to 0 if the diff operation is meaningless).
    *GETBUF and *DIFFBUF must be freed by the caller.  */
 char const *
-version_controller (char const *filename, int readonly,
+version_controller (char const *filename, bool readonly,
 		    struct stat const *filestat, char **getbuf, char **diffbuf)
 {
   struct stat cstat;
   char const *filebase = base_name (filename);
   char const *dotslash = *filename == '-' ? "./" : "";
-  size_t dir_len = filebase - filename;
+  size_t dirlen = filebase - filename;
   size_t filenamelen = strlen (filename);
   size_t maxfixlen = sizeof "SCCS/" - 1 + sizeof SCCSPREFIX - 1;
   size_t maxtrysize = filenamelen + maxfixlen + 1;
@@ -285,8 +290,8 @@ version_controller (char const *filename, int readonly,
 
   strcpy (trybuf, filename);
 
-#define try1(f,a1)    (sprintf (trybuf + dir_len, f, a1),    stat (trybuf, &cstat) == 0)
-#define try2(f,a1,a2) (sprintf (trybuf + dir_len, f, a1,a2), stat (trybuf, &cstat) == 0)
+#define try1(f,a1)    (sprintf (trybuf + dirlen, f, a1),    stat (trybuf, &cstat) == 0)
+#define try2(f,a1,a2) (sprintf (trybuf + dirlen, f, a1,a2), stat (trybuf, &cstat) == 0)
 
   /* Check that RCS file is not working file.
      Some hosts don't report file name length errors.  */
@@ -362,6 +367,23 @@ version_controller (char const *filename, int readonly,
 	*diffbuf = 0;
 
       r = "ClearCase";
+     }
+  else if (!readonly && filestat &&
+           (getenv("P4PORT") || getenv("P4USER") || getenv("P4CONFIG")))
+    {
+      if (getbuf)
+	{
+	  char *p = *getbuf = xmalloc (maxgetsize);
+	  strcpy (p, PERFORCE_CO);
+	  p += sizeof PERFORCE_CO - 1;
+	  p += quote_system_arg (p, filename);
+	  *p = '\0';
+	}
+
+      if (diffbuf)
+	*diffbuf = 0;
+
+      r = "Perforce";
     }
 
   free (trybuf);
@@ -369,12 +391,12 @@ version_controller (char const *filename, int readonly,
 }
 
 /* Get FILENAME from version control system CS.  The file already exists if
-   EXISTS is nonzero.  Only readonly access is needed if READONLY is nonzero.
+   EXISTS.  Only readonly access is needed if READONLY.
    Use the command GETBUF to actually get the named file.
    Store the resulting file status into *FILESTAT.
-   Return nonzero if successful.  */
-int
-version_get (char const *filename, char const *cs, int exists, int readonly,
+   Return true if successful.  */
+bool
+version_get (char const *filename, char const *cs, bool exists, bool readonly,
 	     char const *getbuf, struct stat *filestat)
 {
   if (patch_get < 0)
@@ -460,7 +482,7 @@ format_linenum (char numbuf[LINENUM_LENGTH_BOUND + 1], LINENUM n)
 	*--p = '0' + (int) (n % 10);
       while ((n /= 10) != 0);
     }
-	   
+
   return p;
 }
 
@@ -490,7 +512,7 @@ fatal (char const *format, ...)
 {
   va_list args;
   fprintf (stderr, "%s: **** ", program_name);
-  vararg_start (args, format);
+  va_start (args, format);
   vfprintf (stderr, format, args);
   va_end (args);
   putc ('\n', stderr);
@@ -524,7 +546,7 @@ pfatal (char const *format, ...)
   int errnum = errno;
   va_list args;
   fprintf (stderr, "%s: **** ", program_name);
-  vararg_start (args, format);
+  va_start (args, format);
   vfprintf (stderr, format, args);
   va_end (args);
   fflush (stderr); /* perror bypasses stdio on some hosts.  */
@@ -540,7 +562,7 @@ void
 say (char const *format, ...)
 {
   va_list args;
-  vararg_start (args, format);
+  va_start (args, format);
   vfprintf (stdout, format, args);
   va_end (args);
   fflush (stdout);
@@ -555,7 +577,7 @@ ask (char const *format, ...)
   int r;
   va_list args;
 
-  vararg_start (args, format);
+  va_start (args, format);
   vfprintf (stdout, format, args);
   va_end (args);
   fflush (stdout);
@@ -566,7 +588,8 @@ ask (char const *format, ...)
 	 since it's unlikely that stdout will be seen by the tty user.
 	 The isatty test also works around a bug in GNU Emacs 19.34 under Linux
 	 which makes a call-process `patch' hang when it reads from /dev/tty.
-	 POSIX.2 requires that we read /dev/tty, though.  */
+	 POSIX.1-2001 XCU line 26599 requires that we read /dev/tty,
+	 though.  */
       ttyfd = (posixly_correct || isatty (STDOUT_FILENO)
 	       ? open (TTY_DEVICE, O_RDONLY)
 	       : -1);
@@ -607,15 +630,15 @@ ask (char const *format, ...)
 
 /* Return nonzero if it OK to reverse a patch.  */
 
-int
+bool
 ok_to_reverse (char const *format, ...)
 {
-  int r = 0;
+  bool r = false;
 
   if (noreverse || ! (force && verbosity == SILENT))
     {
       va_list args;
-      vararg_start (args, format);
+      va_start (args, format);
       vfprintf (stdout, format, args);
       va_end (args);
     }
@@ -623,19 +646,17 @@ ok_to_reverse (char const *format, ...)
   if (noreverse)
     {
       printf ("  Skipping patch.\n");
-      skip_rest_of_patch = TRUE;
-      r = 0;
+      skip_rest_of_patch = true;
     }
   else if (force)
     {
       if (verbosity != SILENT)
 	printf ("  Applying it anyway.\n");
-      r = 0;
     }
   else if (batch)
     {
       say (reverse ? "  Ignoring -R.\n" : "  Assuming -R.\n");
-      r = 1;
+      r = true;
     }
   else
     {
@@ -648,7 +669,7 @@ ok_to_reverse (char const *format, ...)
 	    {
 	      if (verbosity != SILENT)
 		say ("Skipping patch.\n");
-	      skip_rest_of_patch = TRUE;
+	      skip_rest_of_patch = true;
 	    }
 	}
     }
@@ -711,7 +732,7 @@ static sigset_t initial_signal_mask;
 static sigset_t signals_to_block;
 
 #if ! HAVE_SIGACTION
-static RETSIGTYPE fatal_exit_handler PARAMS ((int)) __attribute__ ((noreturn));
+static RETSIGTYPE fatal_exit_handler (int) __attribute__ ((noreturn));
 static RETSIGTYPE
 fatal_exit_handler (int sig)
 {
@@ -721,7 +742,7 @@ fatal_exit_handler (int sig)
 #endif
 
 void
-set_signals (int reset)
+set_signals (bool reset)
 {
   int i;
 #if HAVE_SIGACTION
@@ -743,7 +764,7 @@ set_signals (int reset)
       sigemptyset (&signals_to_block);
       for (i = 0;  i < NUM_SIGS;  i++)
 	{
-	  int ignoring_signal;
+	  bool ignoring_signal;
 #if HAVE_SIGACTION
 	  if (sigaction (sigs[i], (struct sigaction *) 0, &initial_act) != 0)
 	    continue;
@@ -939,7 +960,13 @@ fetchname (char *at, int strip_leading, time_t *pstamp)
 	  }
 	else if (ISSPACE ((unsigned char) *t))
 	  {
+	    /* Allow file names with internal spaces,
+	       but only if a tab separates the file name from the date.  */
 	    char const *u = t;
+	    while (*u != '\t' && ISSPACE ((unsigned char) u[1]))
+	      u++;
+	    if (*u != '\t' && strchr (u + 1, '\t'))
+	      continue;
 
 	    if (set_time | set_utc)
 	      stamp = str2time (&u, initial_time,
@@ -994,3 +1021,105 @@ Fseek (FILE *stream, file_offset offset, int ptrname)
   if (file_seek (stream, offset, ptrname) != 0)
     pfatal ("fseek");
 }
+
+typedef struct
+{
+  dev_t		fid_dev;
+  ino_t		fid_ino;
+  time_t	fid_mtime;
+  unsigned long	fid_mtimensec;
+} file_id;
+
+unsigned
+file_id_hasher (file_id *entry, unsigned table_size)
+{
+  return ((unsigned long) entry->fid_ino +
+  	  (unsigned long) entry->fid_dev +
+  	  (unsigned long) entry->fid_mtime +
+	  (unsigned long) entry->fid_mtimensec) % table_size;
+}
+
+bool
+file_id_comparator (file_id *entry1, file_id *entry2)
+{
+  return (entry1->fid_dev	== entry2->fid_dev &&
+	  entry1->fid_ino	== entry2->fid_ino &&
+	  entry1->fid_mtime	== entry2->fid_mtime &&
+	  entry1->fid_mtimensec == entry2->fid_mtimensec);
+}
+
+void
+file_id_freer (file_id *entry)
+{
+  free (entry);
+}
+
+Hash_table *file_id_hash;
+
+/* Check if the file identified by FILENAME and PST was already seen. If the
+   file was already seen, returns TRUE.  If the file has not yet been seen
+   and INSERT is TRUE, it is inserted.  PST or FILENAME may be NULL (but not
+   both of them).  */
+
+static bool
+fid_search (const char *filename, const struct stat *pst, bool insert)
+{
+  struct stat st;
+
+  if (!file_id_hash)
+    {
+      file_id_hash = hash_initialize (0, NULL, (Hash_hasher) file_id_hasher,
+				      (Hash_comparator) file_id_comparator,
+				      (Hash_data_freer) file_id_freer);
+      if (!file_id_hash)
+	pfatal ("hash_initialize");
+    }
+
+  if (!pst)
+    {
+      if (stat (filename, &st) != 0)
+	pfatal ("%s", quotearg (filename));
+      pst = &st;
+    }
+
+  if (insert)
+    {
+      file_id *pfid = xmalloc (sizeof (file_id)), *old_pfid;
+      pfid->fid_dev	  = pst->st_dev;
+      pfid->fid_ino	  = pst->st_ino;
+      pfid->fid_mtime	  = pst->st_mtime;
+#if defined(HAVE_STAT_NSEC)
+      pfid->fid_mtimensec = pst->st_mtimensec;
+#elif defined(HAVE_STAT_TIMEVAL)
+      pfid->fid_mtimensec = pst->st_mtim.tv_nsec;
+#else
+      pfid->fid_mtimensec = 0;
+#endif
+      old_pfid = hash_insert (file_id_hash, pfid);
+      if (!old_pfid)
+	pfatal ("hash_insert");
+      else if (old_pfid != pfid)
+        {
+	  free (pfid);
+	  return true;
+	}
+      else
+	return false;
+    }
+  else
+    {
+      file_id fid;
+      fid.fid_dev	= pst->st_dev;
+      fid.fid_ino	= pst->st_ino;
+      fid.fid_mtime	= pst->st_mtime;
+#if defined(HAVE_STAT_NSEC)
+      fid.fid_mtimensec = pst->st_mtimensec;
+#elif defined(HAVE_STAT_TIMEVAL)
+      fid.fid_mtimensec = pst->st_mtim.tv_nsec;
+#else
+      fid.fid_mtimensec = 0;
+#endif
+      return hash_lookup (file_id_hash, &fid) != 0;
+    }
+}
+
